# %%
from build123d import *
from ocp_vscode import show, set_port
from ocp_vscode.show import _convert
import json
import orjson
from ocp_vscode.comms import default

# %%

# %%
a, b, thickness, distance, fontsize, depth = 13.6, 8.0, 1.0, 0.3, 20, 2

pts = ((-a, 0), (0, b), (a, 0))

o = Text("O", fontsize, "Futura")
cp = Text("CP", fontsize, "Futura")

l1 = ThreePointArc(*pts)
l1 += ThreePointArc(*(l1 @ 1, (0, -b), l1 @ 0))

l2 = offset([l1], thickness)
l3 = offset([l1], thickness + distance)
l4 = offset([l1], -distance)

eye_face = make_face(l2) - make_face(l1)
eye = extrude(eye_face, -2)

eye_mask_face = make_face(l3) - make_face(l4)
eye_mask = extrude(eye_mask_face, -2)

logo_o = extrude(o, -depth)
logo_cp = Pos(22.5, 0, 0) * extrude(cp, -depth)
logo = logo_o + (logo_cp - eye_mask)

center_wire = Wire.make_wire(logo_o.faces().sort_by(Axis.Z).first.edges()[16:])
center = extrude(make_face(center_wire), 2)

eye = eye + center

logo = Plane.XZ * Pos(0, -20, 0) * logo
eye = Plane.XZ * Pos(0, -20, 0) * eye
# %%
# show(
#     logo,
#     eye,
#     colors=[(85, 160, 227), (51,51,51)],
#     names=["OCP", "Eye"],
#     grid=(True, False, False),
#     ortho=False,
#     axes0=True,
#     zoom=0.8,
#     position=[92.4, -99.3, -6.6],
#     quaternion=[0.6047, 0.2156, 0.2603, 0.7212],
# )

c = _convert(
    logo,
    eye,
    colors=[(85, 160, 227), "#333"],
    names=["OCP", "Eye"],
    grid=(True, False, False),
    ortho=False,
    axes0=True,
    zoom=0.8,
    position=[92.4, -99.3, -6.6],
    quaternion=[0.6047, 0.2156, 0.2603, 0.7212],
)

# %%
with open("logo.txt", "w") as fp:
    fp.write(json.dumps(c[0], separators=(",", ":")))

# %%
with open("mapping.txt", "w") as fp:
    fp.write(orjson.dumps(c[1], default=default).decode("utf-8"))



# %%

import copy
from build123d import *
from ocp_vscode import *
from ocp_vscode.colors import ListedColorMap


def reference(obj, loc):
    return copy.copy(obj).move(loc)


sphere = Sphere(1)
spheres = [reference(sphere, loc) for loc in GridLocations(1, 2, 1, 20)]


# %%
show(*spheres, colors=ColorMap.tab20(alpha=0.8))
# %%
show(*spheres, colors=ColorMap.tab20(reverse=True))
# %%
show(*spheres, colors=ColorMap.segmented(20, "mpl:Greens", alpha=0.8))
# %%
show(*spheres, colors=ColorMap.segmented(10, "mpl:Greens", alpha=0.8))
# %%
show(*spheres, colors=ColorMap.golden_ratio("mpl:Greens", alpha=0.8))
# %%
show(*spheres, colors=ColorMap.golden_ratio("mpl:Greens", reverse=True))
# %%
show(*spheres, colors=ColorMap.segmented(10, "mpl:summer", reverse=True))
# %%
show(*spheres, colors=ColorMap.seeded(42, "mpl:summer"))
# %%
show(*spheres, colors=ColorMap.seeded(4242, "mpl:summer"))
# %%
show(*spheres, colors=ColorMap.segmented(20, "hsv", reverse=False))
# %%
show(*spheres, colors=ColorMap.segmented(10, "hsv", reverse=False))
# %%
show(*spheres, colors=ColorMap.golden_ratio("hsv", alpha=0.8))
# %%
show(*spheres, colors=ColorMap.seeded(42, "hsv"))
# %%
show(
    *spheres,
    colors=ColorMap.seeded(59798267586177, "rgb", lower=10, upper=100, brightness=2.2)
)
# %%

set_colormap(ColorMap.golden_ratio("mpl:Blues", alpha=0.8))
show(*spheres)
# %%
show(*spheres[:10])
# %%
set_colormap(ColorMap.tab20(alpha=0.8))
show(*spheres)
# %%
set_colormap(ColorMap.seeded(42, "hsv", alpha=0.8))
show(*spheres)
# %%
set_colormap(ColorMap.segmented(20, "hsv"))
show(*spheres)
# %%
reset_show()

show_object(spheres[0])
show_object(spheres[1])
show_object(spheres[2])
show_object(spheres[3], options={"color": "black", "alpha": 1.0})
show_object(spheres[4])
show_object(spheres[5])
show_object(spheres[6])

# %%

set_colormap(ListedColorMap(["red", "green", "blue"]))
show(*spheres, colors=[None, "yellow"])
# %%

boxes = [loc * Box(1, 1, 1) for loc in [Pos(2 * i, 0, 0) for i in range(20)]]

colors = ColorMap.listed(20, "mpl:turbo", reverse=False)
show(*boxes, colors=colors)

# %%

colors = ColorMap.listed(colors=["red", "green", "blue"])
show(*boxes, colors=colors)

# %%

sphere = Sphere(1)
show(*spheres, colors=ColorMap.segmented(20, "mpl:hot"))

# %%



# %%
from build123d import *
from ocp_vscode import *

b = Box(1, 2, 3)
b = fillet(b.edges(), 0.2)
c = Pos(2, 2, 0) * Cylinder(0.5, 2)

#
# Use config with show for one time setting configurations
#

# %%
# Use default settings
show(b, c)

# %%
# Unset glass mode
show(b, c, glass=False)

# %%
# Remove tools
show(b, c, tools=False)

# %%
# Reset the above settings
show(b, c, tools=True, glass=True)

# %%
show(b, c, orbit_control=True)

# %%
import cadquery as cq

b1 = cq.Workplane().box(1, 2, 3)
b2 = cq.Workplane().box(1, 2, 1)
b3 = cq.Workplane().box(2, 1, 2)
a1 = cq.Assembly(name="a")
b4 = cq.Assembly(name="b")
b4.add(b2, name="b2", loc=Location((-3, -3, -3)))
b4.add(b3, name="b3", loc=Location((3, 3, 3)))
a1.add(b4)
a1.add(b1, name="b1")

show(a1, collapse=Collapse.LEAVES, debug=True)

# %%
show(a1, collapse=Collapse.ALL, debug=True)

# %%
show(a1, collapse=Collapse.ROOT, debug=True)

# %%
show(a1, collapse=Collapse.NONE, debug=True)

# %%
# Provide names to objects

show(b, c, names=["b", "c"])

# %%
# Provide colors to objects

show(b, c, colors=["red", "green"])

# %%
# Provide tranparency to objects

show(b, c, alphas=[0.9, 0.1])

# %%
# Set the progress indicator to only show instances
show(b, c, progress="-")

# %%
# Show axes at center of overall mass

show(b, c, axes=True, axes0=False)

# %%
# Show axes at (0, 0, 0)

show(b, c, axes=True, axes0=True)

# %%
# Show XY grid

show(b, c, grid=(True, True, False))

# %%
# Show XY grid with more ticks

show(b, c, grid=(True, False, True), ticks=20)

# %%
# Show object in perspective mode

show(b, c, ortho=False)

# %%
# Use Y axis as up direction

show(b, c, up="Y")

# %%
# Show all objects with transparency and default opacity level

show(b, c, transparent=True)

# %%
# Show all objects with transparency and set the opacity level

show(b, c, transparent=True, default_opacity=0.1)

# %%
show(b, c, black_edges=True)

# %%
# Change default color to cyan

show(b, c, default_color=(0, 255, 255))

# %%
# Change default color to red (css color name)

show(b, c, default_color="red")


# %%
# Set the default edge color

show(b, c, default_edgecolor="red")

# %%
# Change position and orientation of camera. Use "status()" to get actual values
show(
    b,
    c,
    zoom=0.3,
    position=[15.4, -7.4, 4.3],
    quaternion=[0.3301, 0.2803, 0.7015, 0.5658],
    target=[4.7, -7.7, -4.3],
)

# %%
# Do not reset the camera

show(b, c, reset_camera=Camera.KEEP, colors=["red", "green"], alphas=[0.3, 0.7])

# %%
show(b, c, reset_camera=Camera.CENTER, colors=["blue", "yellow"], alphas=[0.3, 0.7])

# %%
# Suppress rendering edges

show(b, c, render_edges=False)

# %%
# Render vertes normals

show(b, c, render_normals=True)

# %%
# Increase tessellation accuracy (a factor, defaulting to 0.1)

show(b, c, deviation=0.001, debug=True, timeit=True)

# %%
# Increase angular tolerance for tessellation

show(b, c, angular_tolerance=0.02, debug=True, timeit=True)

# %%
# Increase accuracy for discretizing edges

show(*b.edges(), debug=True, timeit=True)

# %%
show(*b.edges(), edge_accuracy=0.02, debug=True, timeit=True)

# %%
# Make the ambinence brighter

show(b, c, ambient_intensity=0.1)

# %%
# Make the direct lights brighter

show(b, c, direct_intensity=0.2)

# %%
reset_defaults()

# %%
# Set pan speed to very fast
show(b, c, pan_speed=10)

# %%
# Set rotate speed to very slow
show(b, c, rotate_speed=0.1)

# %%
# Set zoom speed to very fast
show(b, c, zoom_speed=10)

# %%
# Enable debug output in python nd javascript
show(b, c, debug=True)

# %%
# Time the tesselllation
show(b, c, timeit=3)

# %%

#
# Dynamically set values
#

show(b, c)

# %%
reset_defaults()

set_viewer_config(
    axes=True,
    axes0=True,
    grid=(True, True, False),
    ortho=False,
    transparent=True,
    black_edges=True,
    zoom=0.5,
    position=[2, -12, 6],
    quaternion=[0.46738, -0.02236, 0.16935, 0.86740],
    target=[0.28, -0.66, -1.90],
    default_edgecolor="red",
    default_opacity=0.2,
    zoom_speed=0.1,
    pan_speed=0.1,
    rotate_speed=0.1,
)

# %%

reset_defaults()

# %%
show(b, c)
set_viewer_config(ambient_intensity=0.1)

# %%
set_viewer_config(direct_intensity=10)

# %%

show(b, c)

# %%

set_viewer_config(tools=False)

# %%

set_viewer_config(tree_width=500, tools=True)

# %%

show(b, c)

# %%

# Dynamically set values
#

show(b, c)

# %%

reset_defaults()

set_defaults(
    axes=True,
    axes0=True,
    grid=(True, True, False),
    ortho=False,
    transparent=True,
    black_edges=False,
    zoom=0.5,
    default_edgecolor="red",
    default_opacity=0.2,
    zoom_speed=0.1,
    pan_speed=0.1,
    rotate_speed=0.1,
    glass=True,
)

# %%

show(b, c, transparent=False, timeit=True)

# %%

show(b, c, default_edgecolor="cyan", axes=False)

# %%

reset_defaults()

# %%
set_defaults(reset_camera=Camera.KEEP, timeit=True)

show(b, c)
transparent = False

# %%
transparent = not transparent
show(b, c, transparent=transparent)

# %%

reset_defaults()
show(b, c)
# %%



# %%
from build123d import *
from ocp_vscode import *

set_defaults(reset_camera=Camera.CENTER)


a = 55 / 2
c100 = Circle(100).edges()[0]

lines = []
vertices = []
for sign in (1, -1):
    c = Vector(sign * 31, 0)
    l = PolarLine(c, sign * 6, sign * a)
    l2 = PolarLine(l @ 1, 100, sign * a - 90)
    vertices.append(c100.find_intersection_points(l2.edge()))
    lines.append(l2)

# combine to a wire and face
wire = Wire.combine(
    [
        Line(lines[0] @ 0, vertices[0]),
        ThreePointArc(vertices[0], Vector(0, -100), vertices[1]),
        Line(vertices[1], lines[1] @ 0),
        Line(lines[1] @ 0, lines[0] @ 0),
    ],
)
face = make_face(wire)
face = fillet(face.vertices().group_by(Axis.Y)[0], 15)

# calculate the centers
circles = face.edges().filter_by(GeomType.CIRCLE).sort_by(Axis.X)
centers = [circles[0].arc_center, Vector(0, -85), circles[2].arc_center]

# extrude and cut holes
disk = Circle(31)
for loc in PolarLocations(31, 6):
    disk += loc * Circle(6)

t = extrude(disk, 16 / 2, both=True)
show(t)
t = fillet(t.edges().filter_by(Axis.Z), 8)
show(t)

t += extrude(face, 7 / 2, both=True)
for center in centers:
    t += Pos(*center) * Cylinder(15, 32)
    t -= Pos(*center) * Cylinder(10, 32)

for loc in PolarLocations(31, 6):
    t -= loc * Hole(3, depth=32)
t -= Hole(20, depth=16)

show(t)

# %%



# %%
from build123d import *
from ocp_vscode import *

set_defaults(reset_camera=Camera.KEEP, clip_planes=True, clip_object_colors=True)

# %%

show(Sphere(0.5))
set_viewer_config(tab="clip")
# %%

p2 = Box(1, 1, 1)

show(
    p2,
    debug=True,
    clip_intersection=False,
    clip_planes=True,
    clip_object_colors=False,
    clip_slider_0=0.1,
    clip_slider_1=0.15,
    clip_slider_2=0.2,
    clip_normal_1=(0.12, -0.36, -0.92),
    clip_normal_2=(-0.58, 0.58, -0.58),
)
set_viewer_config(tab="clip")

# %%

show(Sphere(0.5))
set_viewer_config(tab="clip")

# %%

show(Sphere(0.5))
set_viewer_config(tab="clip", clip_object_colors=True)

# %%
show(Sphere(0.5), reset_camera=True)
set_viewer_config(tab="clip")

# %%

show(
    Box(1, 2, 3, align=(Align.MIN, Align.MIN, Align.MIN)),
    center_grid=True,
    grid=(True, False, False),
)
# %%
set_defaults(show_parent=True, reset_camera=Camera.RESET)
show(Box(1, 2, 3).vertices())
# %%
with BuildPart() as p:
    Box(0.1, 0.1, 2)

a = {
    "a": Vector(1, 2, 3),
    "b": [
        Pos(2, 2, 2) * Cylinder(1, 1),
        (1, 2, 3),
        p,
        p.part,
        {"c": Vector(5, 2, 3), "d": Pos(-3, 0, 0) * Box(1, 2, 3), "e": 123},
    ],
}
x = 0
b = [
    Pos(2, 4, 2) * Sphere(1),
    "wert",
    p,
    p.part,
    {"x": Pos(-5, -5, 0) * Box(2, 1, 0.5), "y": 123},
]
show_all()
# %%



# pylint: disable=missing-class-docstring,missing-function-docstring

import unittest
import pytest
import json
import math
from OCP.BRepPrimAPI import BRepPrimAPI_MakeBox
from OCP.BRepFilletAPI import BRepFilletAPI_MakeFillet
from OCP.gp import gp_Pnt, gp_Vec, gp_Dir
from ocp_vscode.show import _convert
from ocp_vscode.backend import ViewerBackend
from ocp_vscode.comms import default
from ocp_vscode.build123d import *
from ocp_tessellate.ocp_utils import get_edges, get_faces, get_vertices


class DirectApiTestCase(unittest.TestCase):
    def assertTupleAlmostEquals(self, first, second, places, msg=None):
        """Check Tuples"""
        self.assertEqual(len(second), len(first))
        for i, j in zip(second, first):
            self.assertAlmostEqual(i, j, places, msg=msg)

    def assertTupleOfTupleAlmostEquals(self, first, second, places, msg=None):
        """Check Tuples"""
        self.assertEqual(len(second), len(first))
        for i, j in zip(second, first):
            self.assertTupleAlmostEquals(i, j, places, msg=msg)

    def assertVectorAlmostEquals(self, first, second, places, msg=None):
        second_vector = Vector(second)
        self.assertAlmostEqual(first.X, second_vector.X, places, msg=msg)
        self.assertAlmostEqual(first.Y, second_vector.Y, places, msg=msg)
        self.assertAlmostEqual(first.Z, second_vector.Z, places, msg=msg)


class VectorTests(DirectApiTestCase):
    def test_tuple(self):
        vec = Vector(1, 2, 3)
        vec2 = Vector((1, 2, 3))
        self.assertTupleAlmostEquals(vec.to_tuple(), (1, 2, 3), 6)
        self.assertTupleAlmostEquals(vec2.to_tuple(), (1, 2, 3), 6)

    def test_gp_pnt(self):
        vec = Vector(gp_Pnt(1, 2, 3))
        self.assertTupleAlmostEquals(vec.to_tuple(), (1, 2, 3), 6)

    def test_gp_vec(self):
        vec = Vector(gp_Vec(1, 2, 3))
        self.assertTupleAlmostEquals(vec.to_tuple(), (1, 2, 3), 6)

    def test_gp_dir(self):
        vec = Vector(gp_Dir(1, 2, 3))
        norm = math.sqrt(1**2 + 2**2 + 3**2)
        self.assertTupleAlmostEquals(vec.to_tuple(), (1 / norm, 2 / norm, 3 / norm), 6)

    def test_normalized(self):
        vec = Vector(gp_Vec(1, 2, 3)).normalized()
        vec2 = Vector(gp_Dir(1, 2, 3))
        self.assertTupleAlmostEquals(vec.to_tuple(), vec2.to_tuple(), 6)

    def test_angle(self):
        vec = Vector(1, 0, 0)
        self.assertAlmostEqual(vec.get_angle(Vector(0, 1, 0)), 90, 6)
        self.assertAlmostEqual(vec.get_angle(Vector(1, 1, 0)), 45, 6)
        self.assertAlmostEqual(vec.get_angle(Vector(1, -1, 0)), 45, 6)


class AxisTests(DirectApiTestCase):
    def test_axis(self):
        a = Axis()
        self.assertTupleAlmostEquals(a.position.to_tuple(), (0, 0, 0), 6)
        self.assertTupleAlmostEquals(a.direction.to_tuple(), (0, 0, 1), 6)
        b = Axis((1, 2, 3), (4, 5, 6))
        self.assertTupleAlmostEquals(b.position.to_tuple(), (1, 2, 3), 6)
        self.assertTupleAlmostEquals(b.direction.to_tuple(), (4, 5, 6), 6)

    def test_axis_x(self):
        a = Axis.X
        self.assertTupleAlmostEquals(a.position.to_tuple(), (0, 0, 0), 6)
        self.assertTupleAlmostEquals(a.direction.to_tuple(), (1, 0, 0), 6)

    def test_axis_y(self):
        a = Axis.Y
        self.assertTupleAlmostEquals(a.position.to_tuple(), (0, 0, 0), 6)
        self.assertTupleAlmostEquals(a.direction.to_tuple(), (0, 1, 0), 6)

    def test_axis_z(self):
        a = Axis.Z
        self.assertTupleAlmostEquals(a.position.to_tuple(), (0, 0, 0), 6)
        self.assertTupleAlmostEquals(a.direction.to_tuple(), (0, 0, 1), 6)


class PlaneTests(DirectApiTestCase):
    def test_face(self):
        box = BRepPrimAPI_MakeBox(1.0, 2.0, 3.0).Solid()
        fillet = BRepFilletAPI_MakeFillet(box)
        for edge in get_edges(box):
            fillet.Add(0.2, edge)
        fillet.Build()
        box = Solid(fillet.Shape())

        faces = box.faces()
        p = Plane(faces[0])

        self.assertTupleAlmostEquals(p.origin.to_tuple(), (0.0, 1.0, 1.5), 6)
        self.assertTupleAlmostEquals(p.x_dir.to_tuple(), (0.0, 0.0, 1.0), 6)
        self.assertTupleAlmostEquals(p.z_dir.to_tuple(), (-1.0, 0.0, 0.0), 6)

        p = Plane(faces[23])

        self.assertTupleAlmostEquals(p.origin.to_tuple(), (1.0, 1.0, 1.5), 6)
        self.assertTupleAlmostEquals(p.x_dir.to_tuple(), (0.0, 0.0, 1.0), 6)
        self.assertTupleAlmostEquals(p.z_dir.to_tuple(), (1.0, 0.0, 0.0), 6)

        p = Plane(faces[10])

        self.assertTupleAlmostEquals(p.origin.to_tuple(), (0.5, 1.0, 3), 6)
        self.assertTupleAlmostEquals(p.x_dir.to_tuple(), (1.0, 0.0, 0), 6)
        self.assertTupleAlmostEquals(p.z_dir.to_tuple(), (0.0, 0.0, 1.0), 6)

    def test_edge(self):
        p = Plane(Vector(1, 1, 1), z_dir=(0, 1, 0))

        self.assertTupleAlmostEquals(p.origin.to_tuple(), (1, 1, 1), 6)
        self.assertTupleAlmostEquals(p.x_dir.to_tuple(), (0, 0, 1), 6)
        self.assertTupleAlmostEquals(p.z_dir.to_tuple(), (0, 1, 0), 6)


class ShapeListTests(DirectApiTestCase):
    def test_lists(self):
        box = BRepPrimAPI_MakeBox(1.0, 2.0, 3.0).Solid()
        fillet = BRepFilletAPI_MakeFillet(box)
        for edge in get_edges(box):
            fillet.Add(0.2, edge)
        fillet.Build()
        box = Solid(fillet.Shape())

        faces = box.faces()
        edges = box.edges()
        vertices = box.vertices()

        self.assertEqual(len(faces), 26)
        self.assertEqual(len(edges), 48)
        self.assertEqual(len(vertices), 24)

        self.assertEqual(len(box.edges().filter_by("LINE")), 24)
        self.assertEqual(len(box.edges().filter_by("CIRCLE")), 24)
        self.assertEqual(len(box.edges().filter_by("OTHER")), 0)

        self.assertEqual(len(box.faces().filter_by("PLANE")), 6)
        self.assertEqual(len(box.faces().filter_by("CYLINDER")), 12)
        self.assertEqual(len(box.faces().filter_by("SPHERE")), 8)
        self.assertEqual(len(box.faces().filter_by("OTHER")), 0)


class ObjectTests(DirectApiTestCase):
    def test_load(self):
        box = BRepPrimAPI_MakeBox(1.0, 2.0, 3.0).Solid()
        _, mapping = _convert(box)
        data = {"model": mapping}
        j = json.dumps(data, default=default).encode("utf-8")

        backend = ViewerBackend(9999)
        backend.load_model(json.loads(j)["model"])

        self.assertEqual(
            list(backend.model.keys()),
            [
                "/Group/Solid",
                "/Group/Solid/faces/faces_0",
                "/Group/Solid/faces/faces_1",
                "/Group/Solid/faces/faces_2",
                "/Group/Solid/faces/faces_3",
                "/Group/Solid/faces/faces_4",
                "/Group/Solid/faces/faces_5",
                "/Group/Solid/edges/edges_0",
                "/Group/Solid/edges/edges_1",
                "/Group/Solid/edges/edges_2",
                "/Group/Solid/edges/edges_3",
                "/Group/Solid/edges/edges_4",
                "/Group/Solid/edges/edges_5",
                "/Group/Solid/edges/edges_6",
                "/Group/Solid/edges/edges_7",
                "/Group/Solid/edges/edges_8",
                "/Group/Solid/edges/edges_9",
                "/Group/Solid/edges/edges_10",
                "/Group/Solid/edges/edges_11",
                "/Group/Solid/vertices/vertices0",
                "/Group/Solid/vertices/vertices1",
                "/Group/Solid/vertices/vertices2",
                "/Group/Solid/vertices/vertices3",
                "/Group/Solid/vertices/vertices4",
                "/Group/Solid/vertices/vertices5",
                "/Group/Solid/vertices/vertices6",
                "/Group/Solid/vertices/vertices7",
            ],
        )
        self.assertEqual(
            [isinstance(v, Face) for k, v in backend.model.items()],
            [
                False,
                True,
                True,
                True,
                True,
                True,
                True,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
        )
        self.assertEqual(
            [isinstance(v, Edge) for k, v in backend.model.items()],
            [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
        )
        self.assertEqual(
            [isinstance(v, Vertex) for k, v in backend.model.items()],
            [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
            ],
        )
        self.assertEqual(
            [isinstance(v, Compound) for k, v in backend.model.items()],
            [
                True,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
        )
        self.assertEqual(
            [v.area for k, v in backend.model.items() if isinstance(v, Face)],
            [6.0, 6.0, 3.0, 3.0, 2.0, 2.0],
        )
        self.assertEqual(
            [v.length for k, v in backend.model.items() if isinstance(v, Edge)],
            [3.0, 2.0, 3.0, 2.0, 3.0, 2.0, 3.0, 2.0, 1.0, 1.0, 1.0, 1.0],
        )
        self.assertEqual(backend.model["/Group/Solid"].volume, 6)

    def test_fillet(self):
        box = BRepPrimAPI_MakeBox(1.0, 2.0, 3.0).Solid()
        fillet = BRepFilletAPI_MakeFillet(box)
        for edge in get_edges(box):
            fillet.Add(0.2, edge)
        fillet.Build()
        box = fillet.Shape()
        _, mapping = _convert(box)
        data = {"model": mapping}
        j = json.dumps(data, default=default).encode("utf-8")

        backend = ViewerBackend(9999)
        backend.load_model(json.loads(j)["model"])

        self.assertEqual(
            [v.geom_type for k, v in backend.model.items()],
            [
                "SOLID",
                "PLANE",
                "CYLINDER",
                "CYLINDER",
                "CYLINDER",
                "CYLINDER",
                "SPHERE",
                "SPHERE",
                "PLANE",
                "PLANE",
                "SPHERE",
                "PLANE",
                "SPHERE",
                "PLANE",
                "CYLINDER",
                "CYLINDER",
                "CYLINDER",
                "CYLINDER",
                "CYLINDER",
                "CYLINDER",
                "CYLINDER",
                "CYLINDER",
                "SPHERE",
                "SPHERE",
                "PLANE",
                "SPHERE",
                "SPHERE",
                "LINE",
                "LINE",
                "LINE",
                "LINE",
                "CIRCLE",
                "CIRCLE",
                "LINE",
                "CIRCLE",
                "LINE",
                "CIRCLE",
                "CIRCLE",
                "LINE",
                "CIRCLE",
                "CIRCLE",
                "CIRCLE",
                "LINE",
                "OTHER",
                "CIRCLE",
                "OTHER",
                "CIRCLE",
                "LINE",
                "LINE",
                "LINE",
                "LINE",
                "LINE",
                "LINE",
                "OTHER",
                "CIRCLE",
                "LINE",
                "LINE",
                "LINE",
                "OTHER",
                "CIRCLE",
                "LINE",
                "LINE",
                "LINE",
                "CIRCLE",
                "CIRCLE",
                "LINE",
                "CIRCLE",
                "CIRCLE",
                "CIRCLE",
                "CIRCLE",
                "LINE",
                "CIRCLE",
                "CIRCLE",
                "CIRCLE",
                "LINE",
                "CIRCLE",
                "LINE",
                "CIRCLE",
                "CIRCLE",
                "OTHER",
                "OTHER",
                "OTHER",
                "OTHER",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
                "VERTEX",
            ],
        )
        vec = [v.to_tuple() for v in backend.model.values() if isinstance(v, Vertex)]
        self.assertTupleOfTupleAlmostEquals(
            vec,
            [
                (-2.7755575615628914e-17, 0.19999999999999998, 0.19999999999999998),
                (-2.7755575615628914e-17, 0.19999999999999998, 2.8),
                (-2.7755575615628914e-17, 1.8, 0.19999999999999998),
                (-2.7755575615628914e-17, 1.8, 2.8),
                (0.19999999999999998, -2.7755575615628914e-17, 0.19999999999999998),
                (0.19999999999999998, -2.7755575615628914e-17, 2.8),
                (0.19999999999999998, 0.19999999999999998, 0.0),
                (0.19999999999999998, 1.8, 0.0),
                (0.19999999999999998, 0.19999999999999998, 3.0),
                (0.19999999999999998, 1.8, 3.0),
                (0.19999999999999998, 2.0, 0.19999999999999998),
                (0.19999999999999998, 2.0, 2.8),
                (0.8, -2.7755575615628914e-17, 0.19999999999999998),
                (0.8, -2.7755575615628914e-17, 2.8),
                (0.8, 0.19999999999999998, 0.0),
                (0.8, 1.8, 0.0),
                (0.8, 0.19999999999999998, 3.0),
                (0.8, 1.8, 3.0),
                (0.8, 2.0, 0.19999999999999998),
                (0.8, 2.0, 2.8),
                (1.0, 0.19999999999999998, 0.19999999999999998),
                (1.0, 0.19999999999999998, 2.8),
                (1.0, 1.8, 0.19999999999999998),
                (1.0, 1.8, 2.8),
            ],
            6,
        )

        vec = [
            (v.radius if v.geom_type in ["CIRCLE"] else -1, v.length)
            for v in backend.model.values()
            if isinstance(v, Edge)
        ]
        self.assertTupleOfTupleAlmostEquals(
            vec,
            [
                (-1, 2.5999999999999996),
                (-1, 1.6),
                (-1, 1.6),
                (-1, 2.5999999999999996),
                (0.2, 0.3141592653589793),
                (0.2, 0.3141592653589793),
                (-1, 2.5999999999999996),
                (0.2, 0.3141592653589793),
                (-1, 1.6),
                (0.2, 0.3141592653589793),
                (0.2, 0.3141592653589793),
                (-1, 1.6),
                (0.2, 0.3141592653589793),
                (0.2, 0.3141592653589793),
                (0.2, 0.3141592653589793),
                (-1, 2.5999999999999996),
                (-1, 1.92367069372179e-17),
                (0.2, 0.3141592653589793),
                (-1, 1.92367069372179e-17),
                (0.2, 0.3141592653589793),
                (-1, 0.6000000000000001),
                (-1, 0.6000000000000001),
                (-1, 2.5999999999999996),
                (-1, 0.6000000000000001),
                (-1, 0.6000000000000001),
                (-1, 1.6),
                (-1, 1.92367069372179e-17),
                (0.2, 0.3141592653589793),
                (-1, 0.6000000000000001),
                (-1, 0.6000000000000001),
                (-1, 1.6),
                (-1, 1.92367069372179e-17),
                (0.2, 0.3141592653589793),
                (-1, 0.6000000000000001),
                (-1, 0.6000000000000001),
                (-1, 2.5999999999999996),
                (0.2, 0.3141592653589793),
                (0.2, 0.3141592653589793),
                (-1, 2.5999999999999996),
                (0.2, 0.3141592653589793),
                (0.2, 0.3141592653589793),
                (0.2, 0.3141592653589793),
                (0.2, 0.3141592653589793),
                (-1, 1.6),
                (0.2, 0.3141592653589793),
                (0.2, 0.3141592653589793),
                (0.2, 0.3141592653589793),
                (-1, 1.6),
                (0.2, 0.3141592653589793),
                (-1, 2.5999999999999996),
                (0.2, 0.3141592653589793),
                (0.2, 0.3141592653589793),
                (-1, 1.92367069372179e-17),
                (-1, 1.92367069372179e-17),
                (-1, 1.92367069372179e-17),
                (-1, 1.92367069372179e-17),
            ],
            6,
        )
        # vec = [
        #     (, v.length, v.width, v.area)
        #     for v in backend.model.values()
        #     if isinstance(v, Face)
        # ]



# %%
from build123d import *
from ocp_vscode import *

import cadquery as cq

from ocp_tessellate.utils import Color


# %% Example "box1"

b = cq.Workplane().box(1, 1, 1)
show(b)


# %% Example "Image Face"

f = ImageFace("examples/object-160x160mm.png", 600 / 912, name="imageplane")
show(f)


# %% Example "orientation box"

b = Box(50, 100, 150)
f = b.faces().sort_by(Axis.Y)[0]
b += extrude(Plane(f) * Rot(0, 0, 0) * Text("ZX / Front", 12), 2)
f = b.faces().sort_by(Axis.Y)[-1]
b += extrude(Plane(f) * Rot(0, 0, 0) * Text("XZ / Back", 12), 2)

f = b.faces().sort_by(Axis.X)[0]
b += extrude(Plane(f) * Rot(0, 0, -90) * Text("ZY / Left", 12), 2)
f = b.faces().sort_by(Axis.X)[-1]
b += extrude(Plane(f) * Rot(0, 0, -90) * Text("YZ / Right", 12), 2)

f = b.faces().sort_by(Axis.Z)[0]
b += extrude(Plane(f) * Rot(0, 0, 90) * Text("XY / Bottom", 12), 2)
f = b.faces().sort_by(Axis.Z)[-1]
b += extrude(Plane(f) * Rot(0, 0, 90) * Text("YX / Top", 12), 2)

show(b)


# %% Example "dir box"

fs = 6
ex = -0.9
b = Box(10, 10, 10)
b = fillet(b.edges().group_by()[-1], 1)
b = chamfer(b.edges().group_by()[0], 1)

f = b.faces().sort_by(Axis.Y)[0]
b -= extrude(Plane(f) * Rot(0, 0, -90) * Text("-Y", fs), ex)
f = b.faces().sort_by(Axis.Y)[-1]
b -= extrude(Plane(f) * Rot(0, 0, -90) * Text("+Y", fs), ex)

f = b.faces().sort_by(Axis.X)[0]
b -= extrude(Plane(f) * Rot(0, 0, -90) * Text("-X", fs), ex)
f = b.faces().sort_by(Axis.X)[-1]
b -= extrude(Plane(f) * Rot(0, 0, -90) * Text("+X", fs), ex)

f = b.faces().sort_by(Axis.Z)[0]
b -= extrude(Plane(f) * Rot(0, 0, 90) * Text("-Z", fs), ex)
f = b.faces().sort_by(Axis.Z)[-1]
b -= extrude(Plane(f) * Rot(0, 0, 90) * Text("+Z", fs), ex)

show(b)


# %% Example Boxes

box1 = cq.Workplane("XY").box(10, 20, 30).edges(">X or <X").chamfer(2)
box1.name = "box1"

box2 = cq.Workplane("XY").box(8, 18, 28).edges(">X or <X").chamfer(2)
box2.name = "box2"

box3 = (
    cq.Workplane("XY")
    .transformed(offset=(0, 15, 7))
    .box(30, 20, 6)
    .edges(">Z")
    .fillet(3)
)
box3.name = "box3"

box4 = box3.mirror("XY").translate((0, -5, 0))
box4.name = "box4"

box1 = box1.cut(box2).cut(box3).cut(box4)

a1 = (
    cq.Assembly(name="ensemble")
    .add(box1, name="red box", color=Color("#d7191c", 0.5))
    .add(box3, name="green box", color=Color("#abdda4"))
    .add(box4, name="blue box", color=Color((43, 131, 186)))
)

show(a1)


# %% Example assembly

s1 = Solid.make_box(1, 1, 1).move(Location((3, 3, 3)))
s1.label, s1.color = "box", "red"

s2 = Solid.make_cone(2, 1, 2).move(Location((-3, 3, 3)))
s2.label, s2.color = "cone", "green"

s3 = Solid.make_cylinder(1, 2).move(Location((-3, -3, 3)))
s3.label, s3.color = "cylinder", "blue"

s4 = Solid.make_sphere(2).move(Location((3, 3, -3)))
s4.label = "sphere"

s5 = Solid.make_torus(3, 1).move(Location((-3, 3, -3)))
s5.label, s5.color = "torus", "cyan"

c2 = Compound(label="c2", children=[s2, s3])
c3 = Compound(label="c3", children=[s4, s5])
c1 = Compound(label="c1", children=[s1, c2, c3])

show(c1)


# %% Example Profile 4040
profile4040 = import_step("examples/profile4040.step")
profile4040.color = "silver"
profile4040.label = "profile 40x40"
show(profile4040)


# %% Example Torus Knot

from math import sin, cos, pi


def points(t0, t1, samples):
    sa = 30
    return [
        Vector(
            sa * (sin(t / samples) + 2 * sin(2 * t / samples)),
            sa * (cos(t / samples) - 2 * cos(2 * t / samples)),
            sa * (-sin(3 * t / samples)),
        )
        for t in range(int(t0), int(t1 * samples))
    ]


zz = points(0, 2 * pi, 200)

with BuildPart() as p:
    with BuildLine() as l:
        m1 = Spline(zz, periodic=False)
        m2 = Line(m1 @ 1, m1 @ 0)  # prevent broken STEP
    pln = Plane(m1 @ 0, z_dir=m1 % 0)
    with BuildSketch(pln) as s:
        Circle(18)
    sweep(is_frenet=True)
p.color = "#852e00"
show(p)


# %% Example faces

b = box1.faces("not(|Z or |X or |Y)")
b.name = "faces"
box1.name = "box1"
show(box1, b, names=["box", "faces"])


# %% Example edges

b = box1.edges("not(|Z or |X or |Y)")
b.name = "edges"
box1.name = "box1"
show(box1, b, names=["box", "edges"])


# %% Example faces

b = box1.vertices("not(|Z or |X or |Y)")
b.name = "vertices"
box1.name = "box1"
show(box1, b, names=["box", "vertices"])


# %% Example drops

b = Box(1, 2, 3) - Plane.YZ * Cylinder(0.5, 1)
b = fillet(b.edges().filter_by(Axis.X), 0.3)
b = chamfer(b.edges().filter_by(Axis.Y), 0.1)
b.label = "drops"
show(b)


# %% Example Single faces
b = Box(1, 2, 3)
b = fillet(b.edges(), 0.2).faces()
show(b)


# %% Example Single edges
b = Box(1, 2, 3)
b = fillet(b.edges(), 0.2).edges()
show(b)


# %% Example Single vertices
b = Box(1, 2, 3)
b = fillet(b.edges(), 0.2).vertices()
show(b)



# %%
import cadquery as cq
from ocp_vscode import *

set_defaults(
    axes=True,
    transparent=False,
    collapse=Collapse.LEAVES,
    reset_camera=Camera.KEEP,
    grid=(True, True, True),
)
# %%
box = cq.Workplane().box(1, 2, 1).edges().chamfer(0.4)

reset_show()
show_object(box, name="Box", options={"alpha": 0.5})

# %%
sphere = cq.Workplane().sphere(0.6)

show_object(
    sphere,
    # show_object args
    "sphere",
    {"color": (10, 100, 110)},
    # three-cad-viewer args
    ortho=False,
)

# %%
show_object(
    box,
    # Clear stack of objects
    clear=True,
    center_grid=True,
    grid=(True, False, False),
)

# %%

set_defaults(
    axes=False,
    transparent=False,
    collapse=Collapse.ROOT,
    reset_camera=Camera.RESET,
    grid=(False, False, False),
)



# %%
import copy
from functools import reduce

import numpy as np

from build123d import *
from ocp_vscode.animation import Animation, collect_paths
from ocp_vscode import show

thickness = 2
height = 40
width = 65
length = 100
diam = 4
tol = 0.05

# %%

#
# Base and top
#


class Base(Part):
    hinge_x1, hinge_x2 = 0.63, 0.87

    hinges_holes = {
        "right_front": Location((-hinge_x1 * width, -hinge_x1 * length), (0, 0, 195)),
        "right_middle": Location((-hinge_x2 * width, 0), (0, 0, 180)),
        "right_back": Location((-hinge_x1 * width, hinge_x1 * length), (0, 0, 165)),
        "left_front": Location((hinge_x1 * width, -hinge_x1 * length), (0, 0, -15)),
        "left_middle": Location((hinge_x2 * width, 0), (0, 0, 0)),
        "left_back": Location((hinge_x1 * width, hinge_x1 * length), (0, 0, 15)),
    }

    stand_holes = {
        "front_stand": Location((0, -0.8 * length), (0, 0, 180)),
        "back_stand": Location((0, 0.75 * length), (0, 0, 0)),
    }

    def __init__(self, label):
        base = extrude(Ellipse(width, length), thickness)
        base -= Pos(Y=-length + 5) * Box(2 * width, 20, 3 * thickness)

        for pos in self.hinges_holes.values():
            base -= pos * Cylinder(
                diam / 2 + tol,
                thickness,
                align=(Align.CENTER, Align.CENTER, Align.MIN),
            )

        for pos in self.stand_holes.values():
            base -= pos * Box(width / 2 + 2 * tol, thickness + 2 * tol, 5 * thickness)

        super().__init__(base.wrapped, label=label)

        # Add joints

        for name, edge in self.hinges_holes.items():
            RigidJoint(f"j_{name}", self, edge)

        for name, pos in self.stand_holes.items():
            RigidJoint(f"j_{name}", self, pos * Rot(0, 0, 90))

        center = self.faces().sort_by().last.center_location
        RigidJoint("j_top", self, center * Pos(Z=height + thickness + 2 * tol))
        RigidJoint("j_bottom", self, center)


base = Base("base")
show(base, render_joints=True)

# %%

#
# Stands
#


class Stand(Part):
    def __init__(self, label):
        self.h = 5

        stand = Box(width / 2 + 10, height + 2 * tol, thickness)
        faces = stand.faces().sort_by(Axis.Y)

        t2 = thickness / 2
        w = height / 2 + tol - self.h / 2
        for i in [-1, 1]:
            rect = Pos(0, i * w, t2) * Rectangle(thickness, self.h)
            block = extrude(rect, self.h)

            m = block.edges().group_by()[-1]
            block = chamfer(
                m.sort_by(Axis.Y).first if i == 1 else m.sort_by(Axis.Y).last,
                length=self.h - 2 * tol,
            )

            stand += block

        for plane in [Plane(faces.first), Plane(faces.last)]:
            stand += plane * Box(
                thickness,
                width / 2,
                thickness,
                align=(Align.CENTER, Align.CENTER, Align.MIN),
            )

        super().__init__(stand.wrapped, label=label)

        RigidJoint(
            "j_bottom",
            self,
            self.faces().sort_by(Axis.Y).last.center_location * Rot(0, 180, 0),
        )


stand = Stand("stand")
show(stand, render_joints=True)

# %%

#
# Legs
#


class UpperLeg(Part):
    def __init__(self, label):
        self.l1 = 50
        self.l2 = 80

        leg_hole = Location((self.l2 - 10, 0), (0, 0, 0))

        line = Polyline(
            (0, 0), (0, height / 2), (self.l1, height / 2 - 5), (self.l2, 0)
        )
        line += mirror(line, about=Plane.XZ)
        face = make_face(line)
        upper_leg = extrude(face, thickness, dir=(0, 0, 1))
        upper_leg = fillet(upper_leg.edges().group_by(Axis.X)[-1], radius=4)

        last = upper_leg.edges()
        upper_leg -= leg_hole * Hole(diam / 2 + tol, depth=thickness)
        self.knee_hole = upper_leg.edges().filter_by(GeomType.CIRCLE) - last

        upper_leg += (
            Pos(0, 0, thickness / 2)
            * Rot(90, 0, 0)
            * Cylinder(diam / 2, 2 * (height / 2 + thickness + tol))
        )

        super().__init__(upper_leg.wrapped, label=label)

        RevoluteJoint(
            "j_hinge",
            self,
            axis=-upper_leg.faces().sort_by(Axis.Y).last.center_location.z_axis,
            angular_range=(0, 180),
        )
        RigidJoint(
            "j_knee_front",
            self,
            leg_hole * Rot(180, 0, 0),
        )
        RigidJoint("j_knee_back", self, leg_hole * Pos(0, 0, thickness))


upper_leg = UpperLeg("upper_leg")
show(upper_leg, render_joints=True)

# %%


class LowerLeg(Part):
    def __init__(self, label):
        self.w = 15
        self.l1 = 20
        self.l2 = 120

        leg_hole = Location((self.l1 - 10, 0), (0, 0, 0))

        line = Polyline((0, 0), (self.l1, self.w), (self.l2, 0))
        line += mirror(line, about=Plane.XZ)
        face = make_face(line)
        lower_leg = extrude(face, thickness, dir=(0, 0, 1))
        lower_leg = fillet(lower_leg.edges().filter_by(Axis.Z), radius=4)

        lower_leg -= leg_hole * Hole(diam / 2 + tol, depth=thickness)

        super().__init__(lower_leg.wrapped, label=label)

        RevoluteJoint(
            "j_front",
            self,
            axis=(leg_hole * Pos(0, 0, thickness) * Rot(0, 180, 0)).z_axis,
            angular_range=(-180, 180),
        )
        RevoluteJoint(
            "j_back",
            self,
            axis=(leg_hole * Pos(0, 0, 0)).z_axis,
            angular_range=(-180, 180),
        )


lower_leg = LowerLeg("lower_leg")
show(lower_leg, render_joints=True)

# %%


def reference(obj, label=None, color=None, location=None):
    new_obj = copy.copy(obj)
    if label is not None:
        new_obj.label = label
    if color is not None:
        new_obj.color = color
    if location is None:
        return new_obj
    else:
        return new_obj.move(location)


class Assembly(Compound):
    def __init__(self, label, children, location=None):
        super().__init__(label=label, children=children)
        # The assembly layer will not contain a part, it only needs a location
        self.location = Location() if location is None else location

    def find_object(self, path):
        def _find(obj, path):
            labels = [child.label for child in obj.children]
            top, _, rest = path.lstrip("/").partition("/")

            try:
                ind = labels.index(top)
            except:
                raise RuntimeError(f"Sub path '{path}' is not valid")

            if rest == "":
                return obj.children[ind]
            else:
                return _find(obj.children[ind], rest)

        name, _, rest = path.strip("/").partition("/")
        if name != self.label:
            raise ValueError(f"Path '{path}' not valid")
        elif rest == "":
            return self

        return _find(self, rest)

    def joint_location(self, obj, joint, **kwargs):
        if isinstance(joint, RevoluteJoint):
            loc = obj.location * joint.relative_axis.location
        elif isinstance(joint, RigidJoint):
            loc = obj.location * joint.relative_location
        else:
            raise NotImplementedError()

        if kwargs.get("angle") is not None:
            loc = loc * Rot(0, 0, kwargs["angle"])

        return loc

    def assemble(
        self, path, joint_name, to_path, to_joint_name, animate=False, **kwargs
    ):
        def _relocate(obj, loc):
            if not isinstance(obj, Assembly):
                obj.location = loc.inverse() * obj.location
            for child in obj.children:
                _relocate(child, loc)

        obj = self.find_object(path)
        joint = obj.joints[joint_name]
        loc = self.joint_location(obj, joint, **kwargs)

        to_obj = self.find_object(to_path)
        to_joint = to_obj.joints[to_joint_name]
        to_loc = self.joint_location(to_obj, to_joint, **kwargs)

        if kwargs.get("angle") is not None:
            to_loc = to_loc * Rot(0, 0, kwargs["angle"])

        if animate:
            # relocate to ensure that the joint is used as origin for the animation
            _relocate(obj, loc)

            # Get the parent Assembly layer
            parent_path, _, _ = path.rpartition("/")

            # For sub assemblies, the location gets pushed one level up to the assembly layer
            parent_obj = self.find_object(parent_path)
            parent_obj.location = to_loc
        else:
            # else just use connect_to for flat assemblies
            to_joint.connect_to(joint, **kwargs)


# %%


base = Base("base")
stand = Stand("stand")
upper_leg = UpperLeg("upper_leg")
lower_leg = LowerLeg("lower_leg")

legs = [
    Assembly(
        children=[
            reference(
                upper_leg,
                label="upper_leg",
                location=Pos(i * 200, j * 50, 0),
            ),
            Assembly(
                children=[
                    reference(
                        lower_leg,
                        label="lower_leg",
                        location=Pos(i * 200 + 80, j * 50, 0),
                    )
                ],
                label=f"{side}_{loc}_lower_leg",
            ),
        ],
        label=f"{side}_{loc}_leg",
    )
    for i, side in enumerate(["left", "right"])
    for j, loc in enumerate(["front", "middle", "back"])
]

hexapod = Assembly(
    children=[
        reference(base, label="bottom", color="grey", location=Pos(-100, -100, 0)),
        reference(base, label="top", color="lightgray", location=Pos(-100, 100, 0)),
        reference(stand, label="front_stand", color="skyblue", location=Pos(0, -60, 0)),
        reference(stand, label="back_stand", color="skyblue", location=Pos(0, -120, 0)),
    ]
    + legs,
    label="hexapod",
)

print(hexapod.show_topology())
show(hexapod, render_joints=True)

# %%

hexapod.assemble("/hexapod/back_stand", "j_bottom", "/hexapod/bottom", "j_back_stand")
hexapod.assemble("/hexapod/front_stand", "j_bottom", "/hexapod/bottom", "j_front_stand")
hexapod.assemble("/hexapod/top", "j_bottom", "/hexapod/bottom", "j_top")

for side in ["left", "right"]:
    for loc in ["front", "middle", "back"]:
        hexapod.assemble(
            f"/hexapod/{side}_{loc}_leg/upper_leg",
            "j_hinge",
            f"/hexapod/bottom",
            f"j_{side}_{loc}",
            angle=-90,
            animate=True,
        )
        hexapod.assemble(
            f"/hexapod/{side}_{loc}_leg/{side}_{loc}_lower_leg/lower_leg",
            "j_front",
            f"/hexapod/{side}_{loc}_leg/upper_leg",
            "j_knee_back" if side == "left" else "j_knee_front",
            angle=-105 if side == "left" else 105,
            animate=True,
        )

print(hexapod.show_topology())
show(hexapod, render_joints=False, debug=False)

# %%

#
# Animation
#


def time_range(end, count):
    return np.linspace(0, end, count + 1)


def vertical(count, end, offset):
    ints = [min(180, (90 + i * (360 // count)) % 360) for i in range(count)]
    heights = [round(20 * np.sin(np.deg2rad(x) - 15), 1) for x in ints]
    heights.append(heights[0])
    return time_range(end, count), heights[offset:] + heights[1 : offset + 1]


def horizontal(end, reverse):
    horizontal_angle = 25
    angle = horizontal_angle if reverse else -horizontal_angle
    return time_range(end, 4), [0, angle, 0, -angle, 0]


animation = Animation(hexapod)

for name in Base.hinges_holes.keys():
    times, values = horizontal(4, "middle" not in name)
    animation.add_track(f"/hexapod/{name}_leg", "rz", times, values)

    times, values = vertical(8, 4, 0 if "middle" in name else 4)
    animation.add_track(f"/hexapod/{name}_leg/{name}_lower_leg", "rz", times, values)

animation.animate(2)

# %%



import orjson
from flask import Flask, render_template
from flask_sock import Sock
from simple_websocket import ConnectionClosed

app = Flask(__name__)
sock = Sock(app)


def config():
    return {"_splash": splash}


def debug_print(*msg):
    print("Debug:", *msg)


def handle_backend(data):
    print("Backend data: " + data)


@app.route("/viewer")
def index():
    return render_template("viewer.html")


python_client = None
javascript_client = None

config = {
    "axes": False,
    "axes0": False,
    "grid": [False, False, False],
    "ortho": True,
    "transparent": False,
    "blackEdges": False,
    "collapse": "R",
    "clipIntersection": False,
    "clipPlaneHelpers": False,
    "clipObjectColors": False,
    "clipNormal0": [-1, 0, 0],
    "clipNormal1": [0, -1, 0],
    "clipNormal2": [0, 0, -1],
    "clipSlider0": -1,
    "clipSlider1": -1,
    "clipSlider2": -1,
    "control": "orbit",
    "up": "Z",
    "ticks": 10,
    "centerGrid": False,
    "position": None,
    "quaternion": None,
    "target": None,
    "measureTools": True,
    "zoom": 1,
    "panSpeed": 0.5,
    "rotateSpeed": 1.0,
    "zoomSpeed": 0.5,
    "timeit": False,
    "default_edgecolor": "#808080",
    "default_color": "#e8b024",
}
status = {}
splash = True


@sock.route("/")
def handle_message(ws):
    global python_client, javascript_client, config, splash

    try:
        while True:
            data = ws.receive()
            if isinstance(data, bytes):
                data = data.decode("utf-8")

            message_type = data[0]
            data = data[2:]
            debug_print("Received data from viewer", message_type, data)

            if message_type == "C":
                python_client = ws
                cmd = orjson.loads(data)
                if cmd == "status":
                    python_client.send(orjson.dumps({"text": status}))
                elif cmd == "config":
                    config["_splash"] = splash
                    python_client.send(orjson.dumps(config))
                elif cmd.type == "screenshot":
                    python_client(orjson.dumps(cmd))

            elif message_type == "D":
                python_client = ws
                debug_print("Received a new model")
                javascript_client.send(data)
                if splash:
                    splash = False

            elif message_type == "U":
                javascript_client = ws
                debug_print("Received incremental UI changes")
                print(data)
                for key, value in orjson.loads(data).items():
                    status[key] = value

            elif message_type == "S":
                python_client = ws
                debug_print("Received a config")
                javascript_client.send(data)
                debug_print("Posted config to view")

            elif message_type == "L":
                javascript_client = ws
                debug_print("Listener registered", data)

            # elif message_type == "B":
            #     handle_backend(data)
            #     debug_print("Model data sent to the backend")

            # elif message_type == "R":
            #     python_client = ws
            #     post_message(data)
            #     debug_print("Backend response received.")

            # if message_type == "C" and not (
            #     isinstance(data, dict) and data["type"] == "screenshot"
            # ):
            #     ws.send(orjson.dumps({}))

    except ConnectionClosed:
        print("Client disconnected")

    except Exception as e:
        print("Error:", e)


if __name__ == "__main__":
    app.run(debug=True)
    sock.init_app(app)



from ocp_tessellate.ocp_utils import deserialize
from asyncio import StreamWriter, StreamReader
import asyncio

BUFFER_SIZE_HEADER = 4  # Length of the header that contains the length of the message


class AsyncServer:
    def __init__(self, host, port):
        self.host = host
        self.port = port

    async def handle_client(self, reader: StreamReader, writer: StreamWriter):
        header = await reader.read(BUFFER_SIZE_HEADER)
        buffer = await reader.read(int.from_bytes(header))

        addr = writer.get_extra_info("peername")
        print(f"Received data from {addr!r}")

        self.handle_data(buffer)

        print("Closing the connection")
        writer.close()

    async def serve(self):
        print(f"Serving on {self.host}:{self.port}")
        server = await asyncio.start_server(self.handle_client, self.host, self.port)

        async with server:
            await server.serve_forever()

    def handle_data(self, data):
        obj = deserialize(data)
        print(obj)


if __name__ == "__main__":
    server = AsyncServer("localhost", 9999)
    asyncio.run(server.serve())



from build123d import *
import pickle
from backend import BUFFER_SIZE_HEADER
import asyncio
from ocp_tessellate.ocp_utils import serialize

box = Box(10, 10, 10).solid()


class AsyncSender:
    def __init__(self, host, port):
        self.host = host
        self.port = port

    async def send(self, obj):
        try:
            data = serialize(obj)
            reader, writer = await asyncio.open_connection(self.host, self.port)

            header = len(data).to_bytes(BUFFER_SIZE_HEADER)
            writer.write(header + data)
            await writer.drain()

            writer.close()
            await writer.wait_closed()

        except Exception as e:
            print(f"Error sending/receiving message: {e}")


if __name__ == "__main__":
    sender = AsyncSender("localhost", 9999)
    asyncio.run(sender.send(box))
